package com.timetec.blesdktestapp.utils;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothManager;
import android.bluetooth.BluetoothProfile;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanResult;
import android.bluetooth.le.ScanSettings;
import android.content.Context;
import android.os.Handler;
import android.support.annotation.NonNull;
import android.util.Log;

import com.timetec.bleandroiddevelopmentkit.TimetecBleSdk;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class BleGateOpener {

    private static final UUID BLE_SERVICE_UUID = UUID.fromString(TimetecBleSdk.BLE_SERVICE_UUID);
    private static final UUID BLE_CONTROL_CHARACTERISTIC_UUID = UUID.fromString(TimetecBleSdk.BLE_CONTROL_CHARACTERISTIC_UUID);
    //The state is considered as "connecting" until the data is sent to the BLE Hardware
    private static boolean connecting, scanning, dataSent;
    private Context mContext;
    private BluetoothDevice bluetoothDevice;
    private BluetoothLeScanner leScanner;
    private BluetoothGatt mBluetoothGattConnected;
    private BluetoothGattCharacteristic mBluetoothGattCharacteristic;
    private TimetecBleSdk timetecBleSdk;
    private Handler mHandler = new Handler();
    private List<ScanFilter> filters = new ArrayList<>();
    private ScanSettings settings;
    private BleGateOpenerCallback callback = null;
    private String targetAddress;
    private int channel;
    private BluetoothGattCallback mGattCallback = new BluetoothGattCallback() {

        @Override
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            Log.d("BleGateOpener: status", Integer.toString(status));
            if (status == BluetoothGatt.GATT_SUCCESS && newState == BluetoothProfile.STATE_CONNECTED) {
                gatt.discoverServices();
            } else if (status == BluetoothGatt.GATT_SUCCESS && newState == BluetoothProfile.STATE_DISCONNECTED) {
                Log.d("Gatt", "Disconnected");
            } else if (!dataSent) {
                if (callback != null) {
                    callback.onFailed();
                }
            }
        }

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            if (status == BluetoothGatt.GATT_SUCCESS) {
                mBluetoothGattCharacteristic = gatt.getService(BLE_SERVICE_UUID).getCharacteristic(BLE_CONTROL_CHARACTERISTIC_UUID);
                sendDataToGatt();
                Log.d("MainActivity", "GAttConnected!");
            }
        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
        }

        @Override
        public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            dataSent = true;
            mBluetoothGattConnected.close();
            mBluetoothGattConnected.disconnect();
        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
        }
    };

    private ScanCallback scanCallback = new ScanCallback() {

        @Override
        public void onScanResult(int callbackType, ScanResult result) {
            if (result != null) {
                Log.d(getClass().getSimpleName(), result.getDevice().getAddress());
                if (result.getDevice().getAddress().equalsIgnoreCase(targetAddress) && !connecting) {
                    bluetoothDevice = result.getDevice();
                    connecting = true;
                    connectDevice();
                }
            }
        }

        @Override
        public void onBatchScanResults(List<ScanResult> results) {
            super.onBatchScanResults(results);
            for (ScanResult r : results) {
                Log.d(getClass().getSimpleName(), r.getDevice().getAddress());
            }

        }

        @Override
        public void onScanFailed(int errorCode) {
            super.onScanFailed(errorCode);
            Log.d("onScanFailed()", "Error Code:" + Integer.toString(errorCode));
        }
    };

    //Constructor
    public BleGateOpener(@NonNull Context context, @NonNull String address) {
        this.mContext = context;
        this.targetAddress = address;

        if (validateMacAddressFormat(address)) {
            timetecBleSdk = new TimetecBleSdk(mContext);
            timetecBleSdk.setMacAddress(targetAddress);
            connecting = false;
            scanning = false;
        } else {
            //Mac Address Invalid
        }
    }

    public BleGateOpener(@NonNull Context context, @NonNull String address, @NonNull BleGateOpenerCallback callback) {
        this.mContext = context;
        this.targetAddress = address;
        this.callback = callback;

        if (validateMacAddressFormat(address)) {
            timetecBleSdk = new TimetecBleSdk(mContext);
            timetecBleSdk.setMacAddress(targetAddress);
            connecting = false;
            scanning = false;
        } else {
            //Mac Address Invalid
        }
    }

    private void scanLeDevice(final boolean enable) {
        if (enable) {
            scanning = true;
            dataSent = false;
            // Stops scanning after a pre-defined scan period.
            mHandler.postDelayed(() -> scanLeDevice(false), 3500);

            Log.d("MainActivity", "Scan Started");
            leScanner.startScan(filters, settings, scanCallback);
        } else {
            Log.d("MainActivity", "Scan Stopped");
            scanning = false;
            leScanner.stopScan(scanCallback);

            //Only consider fail if the scan stops and fail to find a device to connect
            if (callback != null && !connecting && !dataSent) {
                callback.onFailed();
            }
        }
    }

    private void connectDevice() {
        if (bluetoothDevice != null) {
            scanLeDevice(false);
            mBluetoothGattConnected = bluetoothDevice.connectGatt(mContext, true, mGattCallback);
        }
    }

    private void sendDataToGatt() {
        mBluetoothGattCharacteristic.setValue(timetecBleSdk.getParamBle16(channel));
        mBluetoothGattConnected.writeCharacteristic(mBluetoothGattCharacteristic);
        connecting = false;
        if (callback != null) {
            callback.onDataSent();
        }
    }

    //Public methods
    public boolean isBluetoothOn() {
        boolean bluetooth_enabled = false;

        BluetoothManager bluetoothManager = (BluetoothManager) mContext.getSystemService(Context.BLUETOOTH_SERVICE);
        BluetoothAdapter bluetoothAdapter;
        if (bluetoothManager != null) {
            bluetoothAdapter = bluetoothManager.getAdapter();
            bluetooth_enabled = bluetoothAdapter.isEnabled();
            leScanner = bluetoothAdapter.getBluetoothLeScanner();
            settings = new ScanSettings.Builder().setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY)
                    .setReportDelay(0)
                    //.setMatchMode(ScanSettings.MATCH_MODE_AGGRESSIVE)
                    .build();
        }

        if (bluetooth_enabled) {
            return true;
        } else {
            return false;
        }
    }

    public void openGate(int channel) {
        if (callback != null) {
            callback.onScanBegin();
        }
        this.channel = channel;
        scanLeDevice(true);
    }

    public static boolean validateMacAddressFormat(String mac) {
        Pattern p = Pattern.compile("^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$");
        Matcher m = p.matcher(mac);
        return m.find();
    }

    //The custom implementation of callbacks
    public interface BleGateOpenerCallback {
        void onScanBegin();

        void onDataSent();

        void onFailed();
    }
}
